from .artifact import FixOutArtifact
from .helper import ReverseFairness, UnfairModel, clazzes
from .runner import FixOutRunner
from .demos import demo_data


from .fairness import (
    equal_opportunity,
    demographic_parity,
    conditional_accuracy_equality,
    predictive_equality,
    predictive_parity,
    equalized_odds,
    )

__all__ = ['artifact', 'fairness', 'helper', 'runner', 'utils', 'demos', 'interface']


from sklearn import datasets
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

import pandas as pd
import numpy as np

from sklearn.preprocessing import OrdinalEncoder, LabelEncoder

from importlib.resources import files

def importBankData():
    
    #csv_path = files('fixout').joinpath('demos/data/bank.data')
    df = pd.read_csv('fixout/demos/data/bank.data',sep=";", header=0)
    y = df['y'].to_numpy()
    encY = LabelEncoder()
    y = encY.fit_transform(y)
    _df = df.drop(['y'],axis=1)
    dataset = _df.to_numpy()


    enc = OrdinalEncoder(encoded_missing_value=-1)
    X_new = enc.fit_transform(dataset)

    X_train, X_test, y_train, y_test = train_test_split(X_new, y, train_size=0.5, random_state=42)

    model = RandomForestClassifier(n_estimators=10)
    model.fit(X_train, y_train)

    sen_features_indexes = [_df.columns.get_loc(x) for x in ["age","marital","contact"]]

    return model, X_train, X_test, y_train, y_test, _df.columns.values.tolist(), generate_dictionary(enc,sen_features_indexes)


def importCompasData():
    
    #csv_path = files('fixout').joinpath('demos/data/compas.data')
    df = pd.read_csv('fixout/demos/data/compas.data',sep=",", header=0)
    y = df['two_year_recid'].to_numpy()
    _df = df.drop(['two_year_recid','id','name','c_case_number','type_of_assessment','v_type_of_assessment','violent_recid'],axis=1)#,'r_charge_desc','vr_charge_degree','decile_score.1','end','event'],axis=1)
    dataset = _df.to_numpy()

    enc = OrdinalEncoder(encoded_missing_value=-1)
    X_new = enc.fit_transform(dataset)

    X_train, X_test, y_train, y_test = train_test_split(X_new, y, train_size=0.25, random_state=42)

    model = RandomForestClassifier(n_estimators=10)
    model.fit(X_train, y_train)

    sen_features_indexes = [_df.columns.get_loc(x) for x in ["sex","age","race"]]

    return model, X_train, X_test, y_train, y_test, _df.columns.values.tolist(), generate_dictionary(enc,sen_features_indexes)

def importAdultData():
    
    csv_path = files('fixout').joinpath('demos/data/adult.data')
    df = pd.read_csv(csv_path,sep=", ", header=0)
    dataset = df.to_numpy()

    enc = OrdinalEncoder()
    _dataset = enc.fit_transform(dataset)

    X_new = _dataset[:,0:13] # without y
    y = _dataset[:,14]

    X_train, X_test, y_train, y_test = train_test_split(X_new, y, train_size=0.50, random_state=42)

    model = RandomForestClassifier(n_estimators=10)
    model.fit(X_train, y_train)

    return model, X_train, X_test, y_train, y_test, df.columns.values.tolist(), generate_dictionary(enc,[5,8,9])

def generate_dictionary(encoder, sens_feature_indexes):
    dictionary = {}
    for i in sens_feature_indexes:
        dictionary[i] = {}
        for j in range(len(encoder.categories_[i])):
            dictionary[i][j] = encoder.categories_[i][j]
    return dictionary


def importGermanData():

    csv_path = files('fixout').joinpath('demos/data/german.data')
    
    df = pd.read_csv(csv_path,sep=" ", header=0)
    y = df['classification'].to_numpy()
    df = df.drop(['classification'],axis=1)
    dataset = df.to_numpy()

    enc = OrdinalEncoder()
    X_new = enc.fit_transform(dataset)

    X_train, X_test, y_train, y_test = train_test_split(X_new, y, train_size=0.75, random_state=42)

    model = RandomForestClassifier()
    model.fit(X_train, y_train)

    return model, X_train, X_test, y_train, y_test, df.columns.values.tolist(), german_dictionary #generate_dictionary(enc,[19,18,8])


german_dictionary = {
    19 : { 
        0 : "yes",
        1 : "no"
    },
    18 : {
        0 : "none",
        1 : "yes", # registered under the customers name  
    },
    8 : {
        0 : "male divorced",
        1 : "female divorced",
        2 : "male single",
        3 : "male married",
        4 : "female single"
    }
}

if __name__ == '__main__':
    importGermanData()


import pandas as pd
import numpy as np

from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

import json

from scipy.stats import spearmanr
from scipy.stats import pearsonr

import plotly
import plotly.graph_objs as go

import plotly.express as px
colors = px.colors.qualitative.Plotly


def reversed_models_plot(perf_metrics, dictionary=None):

    data = {}

    for sfeature_name, pmetrics in perf_metrics.items(): # for each sensitive feature

        if sfeature_name not in data :
            data[sfeature_name] = []

        data_per_sensF = data[sfeature_name]

        for method_name, metrics in pmetrics: # for each value in the sensitive feature
        
            x = method_name
            y = [m for m in metrics]
            z = ["accuracy","precision","recall"]

            data_per_sensF.append((x,y,z))

    graphJSON_list = []

    for sfeature_name, data_per_sensF in data.items():
        for _x,_y,_z in data_per_sensF:

            data = [
                go.Scatter(
                    x = [_x], 
                    y = [_y[i]],
                    mode = 'markers',
                    name = _z[i],
                    #marker = dict(color = colors[(_z) % len(colors)]),
                    showlegend=False
                ) for i in range(len(_y))
            ]

            graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)
            graphJSON_list.append((sfeature_name, graphJSON))
        
            
    return graphJSON_list


def unfair_model_plot(perf_metrics, dictionary=None):

    graphJSON_list = []

    for method_name, pmetrics in perf_metrics:

        _y = [i for i in pmetrics]
        _z = ["accuracy","precision","recall"]

        data = [
            go.Scatter(
                x = [method_name], 
                y = [_y[i]],
                mode = 'markers',
                name = _z[i],
                #marker = dict(color = colors[(_z) % len(colors)]),
                showlegend=False
            ) for i in range(len(_y))
        ]

        graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)
        graphJSON_list.append(graphJSON)
            
    return graphJSON_list


def generate_visu(X,label,sensitive_feature_index, sensitive_feature_names, dictionary=None):

    if X is None or label is None:
        return None
    
    tsne = TSNE(n_components=2)
    tsne_result = tsne.fit_transform(X)
    tsne_result_df = pd.DataFrame({'tsne_1': tsne_result[:,0], 'tsne_2': tsne_result[:,1], 'label': label})

    pca = PCA(n_components=2)    
    components = pca.fit_transform(X)
    pca_components_df = pd.DataFrame({'component_1': components[:,0], 'component_2': components[:,1], 'label': label})

    graphJSON_list = []
    for i in range(len(sensitive_feature_index)):
        sens_f_index = [i]
        
        data_pca = [
            go.Scatter(
                x=pca_components_df['component_1'],
                y=pca_components_df['component_2'],
                text=pca_components_df['label'],
                mode = 'markers',
                marker = dict(
                    color = pca_components_df['label'],
                    symbol=pd.DataFrame(X)[sens_f_index]
                    ),
                name="",
                showlegend=True
            )
        ]

        data_tsne = [
            go.Scatter(
                x=tsne_result_df['tsne_1'],
                y=tsne_result_df['tsne_2'],
                text=pca_components_df['label'],
                mode = 'markers',
                marker = dict(
                    color = tsne_result_df['label'],
                    symbol=pd.DataFrame(X)[sens_f_index]
                    ),
                name="",
                showlegend=True            
            )
        ]

        graphJSON_pca = json.dumps(data_pca, cls=plotly.utils.PlotlyJSONEncoder)
        graphJSON_tsne = json.dumps(data_tsne, cls=plotly.utils.PlotlyJSONEncoder)
        
        graphJSON_list.append((sensitive_feature_names[i],sens_f_index,graphJSON_pca,graphJSON_tsne))

    return graphJSON_list

# not working yet
def generate_visu3d(X,label,sensitive_feature_index, sensitive_feature_names, dictionary=None):

    if X is None or label is None:
        return None
    
    tsne = TSNE(n_components=3)
    tsne_result = tsne.fit_transform(X)
    tsne_result_df = pd.DataFrame({'tsne_1': tsne_result[:,0], 'tsne_2': tsne_result[:,1], 'tsne_3': tsne_result[:,2], 'label': label})

    pca = PCA(n_components=3)    
    components = pca.fit_transform(X)
    pca_components_df = pd.DataFrame({'component_1': components[:,0], 'component_2': components[:,1], 'component_3': components[:,2], 'label': label})

    graphJSON_list = []
    for i in range(len(sensitive_feature_index)):
        sens_f_index = [i]
        
        data_pca = [
            go.Scatter3d(
                x=pca_components_df['component_1'],
                y=pca_components_df['component_2'],
                z=pca_components_df['component_3'],
                text=pca_components_df['label'],
                mode = 'markers',
                name="",
                showlegend=True
            )
        ]

        data_tsne = [
            go.Scatter3d(
                x=tsne_result_df['tsne_1'],
                y=tsne_result_df['tsne_2'],
                z=tsne_result_df['tsne_3'],
                text=pca_components_df['label'],
                mode = 'markers',
                name="",
                showlegend=True            
            )
        ]

        graphJSON_pca = json.dumps(data_pca, cls=plotly.utils.PlotlyJSONEncoder)
        graphJSON_tsne = json.dumps(data_tsne, cls=plotly.utils.PlotlyJSONEncoder)
        
        graphJSON_list.append((sensitive_feature_names[i],sens_f_index,graphJSON_pca,graphJSON_tsne))

    return graphJSON_list

def correlation_analysis(X,
                         sensitive_feature_index, 
                         sensitive_feature_names, 
                         f_names,
                         func_corr=pearsonr):

    if X is None:
        return (None, None)

    n_features = len(X[0])
    corr_matrix = np.zeros((n_features,n_features))
    for i in range(n_features):
        corr_matrix[i] = [func_corr(X[:,i],X[:,j])[0] for j in range(n_features)]

    #corr_matrix = pearsonr(X,X)#,rowvar=False)
    #corr_matrix1 = spearmanr(X,X)

    data = [
        go.Heatmap(
            z=corr_matrix,
            x=f_names,
            y=f_names
        )
    ]
    
    graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)

    #upper_matrix = np.triu(corr_matrix,k=1) # above the diagonal
    
    rankings = []

    for i in range(len(sensitive_feature_index)):
        index_sens_feature = sensitive_feature_index[i]
        name_sens_feature = sensitive_feature_names[i]
        ranking = corr_matrix[index_sens_feature]
        tuple = []
        for r in range(len(ranking)):
            if f_names[r] != name_sens_feature:
                if abs(ranking[r]) >= 0.1:
                    tuple.append((f_names[r],ranking[r]))

        tuple.sort(reverse=True,key=lambda x: abs(x[1]))
        rankings.append((name_sens_feature,tuple))


    return (graphJSON, rankings)


def statsNonThreshold(X,
                      y,
                      sensitive_feature_index, 
                      sensitive_feature_names, 
                      dictionary = None):

    if X is None or y is None:
        return None

    labels = pd.DataFrame(y)
    possible_labels = list(set([int(x) for x in y]))
    df = pd.DataFrame(np.array(X).astype(int))
    
    graphJSON_dic = []
    for i in range(len(sensitive_feature_names)): # for each sensitive feature
        sens_f_index = sensitive_feature_index[i]
                
        #print(df.iloc[:,sens_f_index].to_numpy().flatten())
        #print(labels)

        for target_label in possible_labels:
            target_indexes =  labels[0] == target_label
            target_indexes_comp =  labels[0] != target_label

            x1 = df.iloc[:,sens_f_index][target_indexes].tolist()
            x2 = df.iloc[:,sens_f_index][target_indexes_comp].tolist()

            if dictionary is not None and sens_f_index in dictionary:
                attrib = dictionary[sens_f_index]
                x1 = [attrib[elem] for elem in x1]
                x2 = [attrib[elem] for elem in x2]

            data = [
                go.Histogram(
                    #x=unprivPop,
                    x= x1,#.to_numpy().flatten(),
                    name="True",
                    showlegend=False,
                    opacity=0.6
                ),
                go.Histogram(
                    #x=unprivPop,
                    x= x2,#.to_numpy().flatten(),
                    name="False",
                    showlegend=False,
                    opacity=0.6
                )
            ]
        
        graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)

        sens_f_name = sensitive_feature_names[i]
        graphJSON_dic.append((sens_f_name, graphJSON))

    return graphJSON_dic

def stats(X,
          y,
          sensitive_feature_index, 
          sensitive_feature_names, 
          sens_f_type, 
          sens_f_unpriv, 
          sens_f_unpriv_original):

    if X is None or y is None:
        return None

    labels = pd.DataFrame(y)
    df = pd.DataFrame(np.array(X).astype(int))
    
    graphJSON_dic = []
    for i in range(len(sensitive_feature_names)): # for each sensitive feature
        sens_f_index = sensitive_feature_index[i]
                
        if sens_f_type[i] == 1: # nonnumeric
            indexes_unpriv =  df.iloc[:,sens_f_index] == sens_f_unpriv[i]
            indexes_priv = df.iloc[:,sens_f_index] != sens_f_unpriv[i]
            legend_unpriv = "=" + str(sens_f_unpriv_original[i])
            legend_priv = "!=" + str(sens_f_unpriv_original[i])
        else: # numeric
            indexes_unpriv = df.iloc[:,sens_f_index] < sens_f_unpriv[i]
            indexes_priv = df.iloc[:,sens_f_index] >= sens_f_unpriv[i]
            legend_unpriv = "<" + str(sens_f_unpriv_original[i])
            legend_priv = ">=" + str(sens_f_unpriv_original[i])

        unprivPop_labels = labels.loc[indexes_unpriv]
        privPop_labels = labels.loc[indexes_priv]

        data = [
            go.Histogram(
                #x=unprivPop,
                x=unprivPop_labels.to_numpy().flatten(),
                name="Unpriv. (" + legend_unpriv + ")",
                showlegend=False
            ),
            go.Histogram(
                #x=privPop,
                x=privPop_labels.to_numpy().flatten(),
                name="Priv. (" + legend_priv + ")",
                showlegend=False
            )
        ]
        
        graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)

        sens_f_name = sensitive_feature_names[i]
        graphJSON_dic.append((sens_f_name, graphJSON))

    return graphJSON_dic

def statsInter(X,
               y,
               sensitive_feature_index, 
               sensitive_feature_names, 
               dictionary = None):

    if X is None or y is None:
        return None

    labels = pd.DataFrame(y)
    possible_labels = list(set([int(x) for x in y]))
    df = pd.DataFrame(np.array(X).astype(int))
    
    graphJSON_dic = []
    for i in range(len(sensitive_feature_names)): # for each sensitive feature

        for j in range(len(sensitive_feature_names)):
            if i > j :

                sens_f_index = sensitive_feature_index[i]
                sens_f_index2 = sensitive_feature_index[j]
                

                for target_label in possible_labels:
                    target_indexes =  labels[0] == target_label
                    target_indexes_comp =  labels[0] != target_label

                    x1a = df.iloc[:,sens_f_index][target_indexes]
                    x1b = df.iloc[:,sens_f_index2][target_indexes]
                    x2a = df.iloc[:,sens_f_index][target_indexes_comp]
                    x2b = df.iloc[:,sens_f_index2][target_indexes_comp]

                    if dictionary is not None and sens_f_index in dictionary and sens_f_index2 in dictionary:
                        attrib = dictionary[sens_f_index]
                        x1a = [attrib[elem] for elem in x1a]
                        x2a = [attrib[elem] for elem in x2a]

                        attrib2 = dictionary[sens_f_index2]
                        x1b = [attrib2[elem] for elem in x1b]
                        x2b = [attrib2[elem] for elem in x2b]

                    data = [
                        go.Histogram(
                            #x=unprivPop,
                            x = [str(x1a[k]) + "-" + str(x1b[k]) for k in range(len(x1a))], # .astype(str),
                            #x= df.iloc[:,inter][target_indexes],#.to_numpy().flatten(),
                            name="True",
                            showlegend=False,
                            opacity=0.6
                        ),
                        go.Histogram(
                            #x=unprivPop,
                            x = [str(x2a[k]) + "-" + str(x2b[k]) for k in range(len(x2a))], #.astype(str),
                            #x= df.iloc[:,inter][target_indexes_comp],#.to_numpy().flatten(),
                            name="False",
                            showlegend=False,
                            opacity=0.6
                        )
                    ]
        
                graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)

                sens_f_name = sensitive_feature_names[i] + " - " + sensitive_feature_names[j]
                graphJSON_dic.append((sens_f_name, graphJSON))

    return graphJSON_dic


import sklearn
from sklearn.ensemble import AdaBoostClassifier,BaggingClassifier,RandomForestClassifier
from fixout.interface.ttypes import ModelType

def getModelType(model):

    if model.__class__ is sklearn.linear_model.LogisticRegression : 
        return ModelType.LR
    elif model.__class__ is AdaBoostClassifier:
        return ModelType.ADA
    elif model.__class__ is BaggingClassifier : 
        return ModelType.BAG
    elif model.__class__ is RandomForestClassifier :
        return ModelType.RF
    
    return None



import os, signal
import sys


from flask import Flask, redirect, render_template, url_for
from flask import request
from flask import jsonify

import plotly
import plotly.graph_objs as go

import pandas as pd
import numpy as np
import json
import pickle

import fixout.web.dataset as dat
import fixout.web.fair_exp as fair_exp

from scipy.stats import spearmanr
from scipy.stats import pearsonr

# Redirect stdout and stderr
#sys.stdout = open(os.devnull, 'w')
#sys.stderr = open(os.devnull, 'w')


import flask.cli
flask.cli.show_server_banner = lambda *args: None

import logging
logging.getLogger("werkzeug").disabled = True

app = Flask(__name__)
#app.config["TEMPLATES_AUTO_RELOAD"] = True

storage_path = "storage/"

html_starter = 'index.html'

class WebData():
    def __init__(self):
        f = open("repport_output.fixout",'rb')
        self.input, self.output = pickle.load(f)

    def data(self):
        return self.input["X"], self.input["y"]
    
    def testing_data(self):
        return self.input["testing_data"]
    
    def meta_data(self):
        f_names = self.input["f_names"] 
        sens_f_index = self.input["sens_f_index"]
        sens_f_names = self.input["sens_f_names"]
        dictionary = self.input["dictionary"]

        return dictionary, f_names, sens_f_index, sens_f_names

    def fairness_metrics(self):
        metrics_list = self.output["metrics_list"]
        other_metrics_list = self.output['nonStandardMetricsToBeCalculated']

        return metrics_list, other_metrics_list


@app.route('/')
def index():

    wdata = WebData()

    dictionary, f_names, sens_f_index, sens_f_names = wdata.meta_data()
    metrics_list, other_metrics_list = wdata.fairness_metrics()

    X, y = wdata.data()
    dataslices = wdata.testing_data()
    

    return render_template(html_starter, 
                           report_details=wdata.input["report_details"],
                           num_data_slices=len(dataslices), 
                           sensitive_features=wdata.input["sens_f_names"],
                           dataslices=dataslices)

@app.route('/d/metrics/<int:data_slice_index>')
def d_metrics(data_slice_index):
    return redirect(url_for('metrics', data_slice_index=data_slice_index, debug=True))

@app.route('/metrics/<int:data_slice_index>')
def metrics(data_slice_index):
    wdata = WebData()
    dictionary, f_names, sens_f_index, sens_f_names = wdata.meta_data()
    if data_slice_index == 0:
        X, y = wdata.data()
    else:
        X, y, _ = wdata.testing_data()[data_slice_index-1]

    full_fairness_plots = fair_exp.generate_fairness_plots(wdata.output["result"],dictionary=dictionary)
    other_fairness_plots = fair_exp.generate_fairness_plots(wdata.output["nonstandardResults"],dictionary=dictionary)

    return render_template("metrics.html",
                           debug=request.args.get('debug', None),
                           sensitive_features=wdata.input["sens_f_names"],
                           full_fairness_plots=full_fairness_plots,
                           other_fairness_plots=other_fairness_plots) # train 


@app.route('/d/visu/<int:data_slice_index>')
def d_visu(data_slice_index):
    return redirect(url_for('visu', data_slice_index=data_slice_index, debug=True))

@app.route('/visu/<int:data_slice_index>')
def visu(data_slice_index):
    wdata = WebData()
    dictionary, f_names, sens_f_index, sens_f_names = wdata.meta_data()
    if data_slice_index == 0:
        X, y = wdata.data()
    else:
        X, y, _ = wdata.testing_data()[data_slice_index-1]

    plots_visu = dat.generate_visu(X, y, sens_f_index, sens_f_names)

    return render_template("visualization.html",
                           debug=request.args.get('debug', None),
                           index_data=data_slice_index,
                           sensitive_features=wdata.input["sens_f_names"],
                           plots_visu=plots_visu) # train 

@app.route('/d/histo/<int:data_slice_index>')
def d_histo(data_slice_index):
    return redirect(url_for('histo', data_slice_index=data_slice_index, debug=True))

@app.route('/histo/<int:data_slice_index>')
def histo(data_slice_index):
    wdata = WebData()
    dictionary, f_names, sens_f_index, sens_f_names = wdata.meta_data()
    if data_slice_index == 0: # indicates training data
        X, y = wdata.data()
    else: # otherwise, testing data is used
        X, y, _ = wdata.testing_data()[data_slice_index-1]

    histograms = dat.statsNonThreshold(X, y, sens_f_index, sens_f_names, dictionary)
    inter_histograms = dat.statsInter(X, y, sens_f_index, sens_f_names, dictionary)

    return render_template("histograms.html",
                           debug=request.args.get('debug', None),
                           index_data=data_slice_index,
                           histograms=histograms,
                           inter_histograms=inter_histograms) 



@app.route('/d/corr/<int:data_slice_index>')
def d_corr(data_slice_index):
    return redirect(url_for('corr', data_slice_index=data_slice_index, debug=True))

@app.route('/corr/<int:data_slice_index>')
def corr(data_slice_index):
    wdata = WebData()
    dictionary, f_names, sens_f_index, sens_f_names = wdata.meta_data()
    if data_slice_index == 0:
        X, _ = wdata.data()
    else:
        X, _, _ = wdata.testing_data()[data_slice_index-1]

    pearCorr_heatmap, pearCorr_rankings = dat.correlation_analysis(X, sens_f_index, sens_f_names, f_names, func_corr=pearsonr)
    speaCorr_heatmap, speaCorr_rankings = dat.correlation_analysis(X, sens_f_index, sens_f_names, f_names, func_corr=spearmanr)

    return render_template("correlation.html",
                           debug=request.args.get('debug', None),
                           sensitive_features=wdata.input["sens_f_names"],
                           index_data=data_slice_index,
                           pearCorr_heatmap=pearCorr_heatmap, 
                           pearCorr_rankings=pearCorr_rankings, 
                           speaCorr_heatmap=speaCorr_heatmap,
                           speaCorr_rankings=speaCorr_rankings) 


@app.route('/d/rev/<int:data_slice_index>')
def d_rev(data_slice_index):
    return redirect(url_for('rev', data_slice_index=data_slice_index, debug=True))

@app.route('/rev/<int:data_slice_index>')
def rev(data_slice_index):
    wdata = WebData()
    dictionary, f_names, sens_f_index, sens_f_names = wdata.meta_data()

    rmodels_perf_plors = dat.reversed_models_plot(perf_metrics=wdata.input["reversed_models"][data_slice_index], dictionary=dictionary)
    unfairmodel_perf_plots = dat.unfair_model_plot(perf_metrics=wdata.input["unfair_model"][data_slice_index], dictionary=dictionary)

    return render_template("reversed.html",
                           debug=request.args.get('debug', None),
                           sensitive_features=wdata.input["sens_f_names"],
                           index_data=data_slice_index,
                           rmodels_perf_plors = rmodels_perf_plors,
                           unfairmodel_perf_plots = unfairmodel_perf_plots) 


#@app.route('/exit')
#def stop():
#    os.kill(os.getpid(), signal.SIGINT)
#    return 

if __name__ == "__main__":
    app.run(debug=False)


from sklearn.exceptions import UndefinedMetricWarning

from sklearn import linear_model
from sklearn import tree
from sklearn import svm
from sklearn.naive_bayes import GaussianNB
from sklearn.ensemble import RandomForestClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import GradientBoostingClassifier

from sklearn.metrics import balanced_accuracy_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score

import warnings
warnings.filterwarnings('ignore')


clazzes = [(linear_model.RidgeClassifier,"linear reg"),
            (tree.DecisionTreeClassifier,"tree"),
            (svm.SVC,"svm"),
            (GaussianNB,"gaussian"),
            (RandomForestClassifier,"rd forest"),
            (MLPClassifier,"neural"),
            (GradientBoostingClassifier,"grad boost")]

class ReverseFairness():
    """
    A class for building and evaluating reversed fairness models. 
    These models attempt to predict sensitive features from target labels.
    """

    def build_reversed_models(self, X, y, sensitivefeatureslist, X_test=None, y_test=None):
        """
        Constructs reversed models to predict sensitive features from target labels.

        Parameters
        ----------
        X : array-like
            Feature matrix (excluding sensitive attributes).
        y : array-like
            Target variable labels.
        sensitivefeatureslist : list
            List of sensitive feature objects, each having a `name` and `featureIndex` attribute.
        X_test : array-like, optional
            Test set feature matrix (default is None).
        y_test : array-like, optional
            Test set target labels (default is None).

        Returns
        -------
        dict
            A dictionary mapping each sensitive feature name to a classifier and its list of calculated performance metrics.
        
        Notes
        -----
        - Assumes `clazzes` is defined externally as a list of (classifier, classifier_name) tuples.
        - Uses `build_reverse_model` to train and evaluate classifiers.
        """
    
        results = {}
        _X_test = None
        _y_test = None

        _y = [[i] for i in y]
        if y_test is not None:
            _y_test = [[i] for i in y_test]
        
        for sens_feature in sensitivefeatureslist:
            
            results[sens_feature.name] = []
            _X = X[:,sens_feature.featureIndex]
            if X_test is not None:
                _X_test = X_test[:,sens_feature.featureIndex]

            for clazz,clazz_name in clazzes:
                r = self.build_reverse_model(clazz, _X, _y, _X_test, _y_test) 
                results[sens_feature.name].append((clazz_name,r))

        return results


    def build_reverse_model(self, clazz, X, y, X_test=None, y_test=None):
        """
        Trains a classifier to predict sensitive attributes based on the target variable.

        Parameters
        ----------
        clazz : Classifier class
            The classifier to be trained (e.g., LogisticRegression, DecisionTree).
        X : array-like
            Training data (sensitive feature values).
        y : array-like
            Target labels.
        X_test : array-like, optional
            Test set sensitive feature values (default is None).
        y_test : array-like, optional
            Test set target labels (default is None).

        Returns
        -------
        tuple
            Calculated performance metrics (balanced accuracy, precision, and recall).
        
        Notes
        -----
        - The classifier is trained with `y` as input and `X` as output (reverse prediction).
        - Calls `eval_perf` to compute performance metrics.
        """
        clf = clazz()
        clf.fit(y, X)

        if X_test is not None and y_test is not None:
            return self.__eval_perf(clf, y_test, X_test)
        else:
            return self.__eval_perf(clf, y, X)

    def __eval_perf(self, clf, X_test, y_test):

        y_pred = clf.predict(X_test)
        
        try:
            baccuracy = balanced_accuracy_score(y_test,y_pred)
            precision = precision_score(y_test, y_pred, average='weighted')
            recall = recall_score(y_test, y_pred, average='weighted')
        except UndefinedMetricWarning:
            print("except UndefinedMetricWarning")

        return baccuracy, precision, recall

    

class UnfairModel():
    """
    A class for building and evaluating unfair models, which predict the target 
    variable using only sensitive attributes. This helps analyze bias in models.
    """

    def build_unfair_model(self, X, y, sensitivefeatureslist, X_test=None, y_test=None):
        """
        Trains a classifier to predict target variable based on sensitive attributes only.

        Parameters
        ----------
        X : array-like
            Training data (sensitive feature values).
        y : array-like
            Target labels.
        sensitivefeatureslist : list
            List of sensitive features
        X_test : array-like, optional
            Test set sensitive feature values (default is None).
        y_test : array-like, optional
            Test set target labels (default is None).

        Returns
        -------
        tuple
            Calculated performance metrics (balanced accuracy, precision, and recall).
        
        """

        results = []

        indexes = [x.featureIndex for x in sensitivefeatureslist]
        _X = X[:,indexes]
        
        for clazz,clazz_name in clazzes:
            
            clf = clazz()
            clf.fit(_X, y)

            if X_test is not None and y_test is not None:
                _X_test = X_test[:,indexes]
                results.append((clazz_name,self.__eval_perf(clf, _X_test, y_test)))
            else:
                results.append((clazz_name,self.__eval_perf(clf, _X, y)))

        return results

    def __eval_perf(self, clf, X_test, y_test):

        y_pred = clf.predict(X_test)
        
        try:
            baccuracy = balanced_accuracy_score(y_test,y_pred)
            precision = precision_score(y_test, y_pred)
            recall = recall_score(y_test, y_pred)
        except UndefinedMetricWarning:
            print("except UndefinedMetricWarning:")


        return baccuracy, precision, recall



import pandas as pd
import numpy as np

import plotly
import plotly.graph_objs as go

import json

import hashlib

import plotly.express as px
colors = px.colors.qualitative.Plotly

fair_metrics_names= ['Equality of opportunity','Demographic parity','Predictive equality','Equalized odds']

def fairness_plots_ensemble(iso_fair_metrics, dictionary=None):

    data_per_sensFeature = {}
    sensFeature_name_index = {}

    for sensitive_feature, calculated_metrics, method_name in iso_fair_metrics: # for each sensitive feature
           
        sensFeature_name_index[sensitive_feature.name] = sensitive_feature.featureIndex

        if sensitive_feature.name not in data_per_sensFeature:
            data_per_sensFeature[sensitive_feature.name] = ([],[],[])
        
        x,y,z = data_per_sensFeature[sensitive_feature.name] 
        
        for possible_value, metrics in calculated_metrics.items(): # for each value in the sensitive feature
            
            for metric in metrics:           
                x.append(fair_metrics_names[metric.type])
                y.append(metric.value)
                z.append(possible_value)

    graphJSON_list = []

    for sensFeature_name, data_sensFeature in  data_per_sensFeature.items():

        _x, _y, _z = data_sensFeature

        for i in range(len(_z)):

            # find the name of the category to be displayed
            _name  = _z[i] if dictionary == None else dictionary[sensFeature_name_index[sensFeature_name]][float(_z[i])]
            
            # obtain un integer in order to define a color later on
            sum_ascii_values = sum([ord(char) for char in _name]) 

            data = [
                go.Scatter(
                    x = [_x[i]], 
                    y = [_y[i]],
                    mode = 'markers',
                    name = _name,
                    marker = dict(size = 10,
                                    color = colors[sum_ascii_values % len(colors)]),
                    showlegend=False
                )
            ]

            graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)
            graphJSON_list.append((sensFeature_name, graphJSON))
            
    return graphJSON_list



def generate_fairness_plots(fair_metrics, dictionary=None):

    data_per_sensFeature = {}
    sensFeature_name_index = {}

    for sensitive_feature, calculated_metrics, method_name in fair_metrics: # for each sensitive feature
           
        sensFeature_name_index[sensitive_feature.name] = sensitive_feature.featureIndex

        if sensitive_feature.name not in data_per_sensFeature:
            data_per_sensFeature[sensitive_feature.name] = {}
        
        data_per_metric = data_per_sensFeature[sensitive_feature.name] 

        for possible_value, metrics in calculated_metrics.items(): # for each value in the sensitive feature
            
            for metric in metrics:
                if metric.type not in data_per_metric:
                    data_per_metric[metric.type] = ([],[],[])
            
                x,y,z = data_per_metric[metric.type]

                x.append(method_name)
                y.append(metric.value)
                z.append(possible_value)

    graphJSON_list = []

    for sensFeature_name, data_sensFeature in  data_per_sensFeature.items():

        graphJSON_list_sensFeature = []

        for metric, data_metric in data_sensFeature.items():
            _x, _y, _z = data_metric

            _names = []

            for i in range(len(_z)):
                # find the name of the category to be displayed
                _name  = _z[i] if dictionary == None else dictionary[sensFeature_name_index[sensFeature_name]][float(_z[i])]
                
                # obtain un integer in order to define a color later on
                sum_ascii_values = sum([ord(char) for char in str(_name)])
                
                _names.append(str(sum_ascii_values))

                

            data = [
                go.Scatter(
                    x = _x, 
                    y = _y,
                    mode = 'markers',
                    #name = _names,
                    #marker = dict(size = 10 if "original" != _x[i] else 15,
                    #                color = colors[sum_ascii_values % len(colors)]),
                    showlegend=False
                )
            ]

            graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)
            graphJSON_list_sensFeature.append((metric, False, graphJSON))
            '''
            for i in range(len(_z)):

                # find the name of the category to be displayed
                _name  = _z[i] if dictionary == None else dictionary[sensFeature_name_index[sensFeature_name]][float(_z[i])]
                
                # obtain un integer in order to define a color later on
                sum_ascii_values = sum([ord(char) for char in str(_name)]) 

                data = [
                    go.Scatter(
                        x = [_x[i]], 
                        y = [_y[i]],
                        mode = 'markers',
                        name = _name,
                        marker = dict(size = 10 if "original" != _x[i] else 15,
                                      color = colors[sum_ascii_values % len(colors)]),
                        showlegend=False
                    )
                ]

                graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)
                graphJSON_list_sensFeature.append((metric, False, graphJSON))
            '''
        
        graphJSON_list.append((sensFeature_name, graphJSON_list_sensFeature))
            
    return graphJSON_list



def full_fairness_metrics2(fair_metrics, num_metrics):

    graphJSON_list = []

    for sensitive_feature, calculated_metrics, method_name in fair_metrics: # for each sensitive feature
        dat = {}
        
        for value, metrics in calculated_metrics.items(): # for each value in the sensitive feature
            
            str_value = str(value)

            for metric in metrics: # for each calculated metric 

                str_metric = str(metric.type)
                if str_metric not in dat:
                    dat[str_metric] = {}
                dat[str_metric][str_value] = metric.value

        for i in range(len(num_metrics)): # for each calculated metric
            
            x_ = []
            y_ = []

            if str(i) in dat:
                for threshold, calc_metric in dat[str(i)].items():
                    x_.append(float(threshold))
                    y_.append(calc_metric)
        
            df = pd.DataFrame({'x': x_, 'y': y_}) # creating a sample dataframe

            data = [
                go.Scatter(
                    x=[method_name], # assign x as the dataframe column 'x'
                    y=df['y'],
                    mode = 'markers',
                    showlegend=False
                )
            ]

            alert_flag = False
            alert_array = abs(df['y']) >= 0.25
            if alert_array.any():
                alert_flag = True

            graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)
            graphJSON_list.append((sensitive_feature.name, str(i), alert_flag, graphJSON))
            
    return graphJSON_list


'''
def create_plot(cursor, report_id, sensitive_features):

    graphJSON_dic = []
    for sensitive_f, name in sensitive_features:

        cursor.execute('SELECT * FROM calculated_metrics WHERE original_model = \'True\' AND reportid_fk = %s AND indexf = %s',(report_id, sensitive_f,))
        metrics_original = cursor.fetchall()

        cursor.execute('SELECT * FROM calculated_metrics WHERE original_model = \'False\' AND reportid_fk = %s AND indexf = %s',(report_id, sensitive_f,))
        metrics_ensemble = cursor.fetchall()

        #print (metrics_original)
        y_original = [metric[2] for metric in metrics_original]
        y_ensemble = [metric[2] for metric in metrics_ensemble]

        x = ['DP','EO','PE']
        df_original = None
        df_ensemble = None
        if (len(y_original) == len(x)) :
            df_original = pd.DataFrame({'x': x, 'y': y_original}) # creating a sample dataframe
        if (len(y_ensemble) == len(x)) :
            df_ensemble = pd.DataFrame({'x': x, 'y': y_ensemble}) # creating a sample dataframe

        data = []
        if df_original is not None:
            data.append(
                go.Bar(
                    name='Original',
                    x=df_original['x'], # assign x as the dataframe column 'x'
                    y=df_original['y']
                )
            )
        if df_ensemble is not None:
            data.append(
                go.Bar(
                    name='FixOut',
                    x=df_ensemble['x'], # assign x as the dataframe column 'x'
                    y=df_ensemble['y']
                )
            )

        graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)
        graphJSON_dic.append((name, graphJSON))

    return graphJSON_dic


def full_fairness_metrics(cursor, reportId, sensitive_features):

    graphJSON_list = []
    alerts = []

    cursor.execute('SELECT DISTINCT metric_name_fk FROM calculated_metrics WHERE reportid_fk = %s',(reportId,))
    metrics = cursor.fetchall()

    for sensitive_f in sensitive_features:
        
        sensitive_f_name = sensitive_f[1]

        for metric_name in metrics:

            alert_flag = False

            cursor.execute('SELECT m_value, threshold_value, report_name FROM calculated_metrics INNER JOIN reports ON reportid_fk = report_id  WHERE reportid_fk = %s AND metric_name_fk = %s AND sensitive_f_name = %s AND original_model = \'True\'',(reportId, metric_name, sensitive_f_name,))
            metrics_original = cursor.fetchall()

            x_original = []
            y_original = []
            for metric in metrics_original:
                y_original.append(metric[0])
                x_original.append(metric[1])

            df_original = pd.DataFrame({'x': x_original, 'y': y_original}) # creating a sample dataframe

            data = [
                go.Bar(
                    #name='Original',
                    x=df_original['x'], # assign x as the dataframe column 'x'
                    y=df_original['y']
                )
            ]

            alert_array = abs(df_original['y']) >= 0.25
            if alert_array.any():
                alert_flag = True

            graphJSON = json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder)
            graphJSON_list.append((sensitive_f_name,metric_name,alert_flag,graphJSON))
            
    return graphJSON_list
'''




from fixout.interface.ttypes import FairMetric, FairMetricEnum

import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.preprocessing import binarize
from sklearn.metrics import confusion_matrix


def equalized_odds(cm0,cm1):
    """
    Calculates the equalized odds (EOD) fairness metric.
    
    Parameters
    ----------
    cm0 : array-like
        Confusion matrix for the unprivileged group.
    cm1 : array-like
        Confusion matrix for the privileged group.

    Returns
    -------
    metric : FairMetric
        The calculated equalized odds metric.
    
    """
    tn0, fp0, fn0, tp0 = cm0 # unprivileged 
    tn1, fp1, fn1, tp1 = cm1 # privileged
    metric = FairMetric()
    metric.type = FairMetricEnum.EOD
    metric.value = ( (tp0/(tp0+fn0)) + (fp0/(fp0+tn0)) ) - ( (tp1/(tp1+fn1)) + (fp1/(fp1+tn1)) )
    return metric

def demographic_parity(cm0,cm1):
    """
    Calculates the demographic parity (DP) fairness metric.
    
    Parameters
    ----------
    cm0 : array-like
        Confusion matrix for the unprivileged group.
    cm1 : array-like
        Confusion matrix for the privileged group.

    Returns
    -------
    metric : FairMetric
        The calculated demographic parity metric.

    """
    tn0, fp0, fn0, tp0 = cm0 # unprivileged 
    tn1, fp1, fn1, tp1 = cm1 # privileged
    metric = FairMetric()
    metric.type = FairMetricEnum.DP
    metric.value = ((tp0+tn0)/(tn0+fp0+fn0+tp0)) - ((tp1+tn1)/(tn1+fp1+fn1+tp1))
    return metric

def equal_opportunity(cm0,cm1):
    """
    Calculates the equal opportunity (EO) fairness metric.
    
    Parameters
    ----------
    cm0 : array-like
        Confusion matrix for the unprivileged group.
    cm1 : array-like
        Confusion matrix for the privileged group.

    Returns
    -------
    metric : FairMetric
        The calculated equal opportunity metric.

    """
    _, _, fn0, tp0 = cm0 # unprivileged 
    _, _, fn1, tp1 = cm1 # privileged
    metric = FairMetric()
    metric.type = FairMetricEnum.EO
    metric.value = (tp0/(tp0+fn0)) - (tp1/(tp1+fn1)) 
    return metric

def predictive_equality(cm0,cm1):
    """
    Calculates the equal predictive equality metric.

    Parameters
    ----------
    cm0 : array-like
        Confusion matrix for the unprivileged group.
    cm1 : array-like
        Confusion matrix for the privileged group.

    Returns
    -------
    metric : FairMetric
        The calculated predictive equality metric.
    """
    _, fp0, _, tp0 = cm0 # unprivileged 
    _, fp1, _, tp1 = cm1 # privileged
    metric = FairMetric()
    metric.type = FairMetricEnum.PE
    metric.value = (fp0/(fp0+tp0)) - (fp1/(fp1+tp1)) 
    return metric

def predictive_parity(cm0,cm1):
    """
    Computes the Predictive Parity (PP) fairness metric.

    Parameters
    ----------
    cm0 : array-like
        Confusion matrix for the unprivileged group.
    cm1 : array-like
        Confusion matrix for the privileged group.

    Returns
    -------
    FairMetric
        The calculated predictive parity metric.
    """
    _, fp0, _, tp0 = cm0 # unprivileged 
    _, fp1, _, tp1 = cm1 # privileged
    metric = FairMetric()
    metric.type = FairMetricEnum.PP
    metric.value = (tp0/(tp0+fp0)) - (tp1/(tp1+fp1))
    return metric

def conditional_accuracy_equality(cm0,cm1):
    """
    Computes the Conditional Accuracy Equality (CEA) fairness metric.

    Parameters
    ----------
    cm0 : array-like
        Confusion matrix for the unprivileged group.
    cm1 : array-like
        Confusion matrix for the privileged group.

    Returns
    -------
    FairMetric
        The calculated conditional accuracy equality metric.

    """
    tn0, fp0, fn0, tp0 = cm0 # unprivileged 
    tn1, fp1, fn1, tp1 = cm1 # privileged
    metric = FairMetric()
    metric.type = FairMetricEnum.CEA
    metric.value = ((tp0/(tp0+fp0))+(tn0/(tn0+fn0))) - ((tp1/(tp1+fp1))+(tn1/(tn1+fn1)))
    return metric



    
def computeFairnessMetrics(metrics, sFeature, X_test, y_test, y_pred):
    """
    Calculate fariness metrics for a given ensemble of instances. 
    
    Supported fariness metrics: Demographic Parity, Equal Opportunity, Predictive Equality.

    Check out the enum types : 'FairMetricEnum'
    
    Parameters
    ----------
    metrics : list
        List of fairness metric types to be calculated.

    sFeature : SensitiveFeature 
        Sensitive feature that will be take into account to calculate fairness metrics.

    X_test : array-like
        Feature matrix of the test set.
    
    y_test : array-like
        True labels of the test set.
    
    y_pred : array-like
        Predicted labels.

    Returns
    -------
    results : list of FairMetric
        List with all requested metrics calculated.
        Returns an empty list if no metric is informed or a problem when calculating the confusion matrix is found.

    """ 
    
    y_test_array = np.array(y_test)
    X_test_array = np.array(X_test)

    results = {}
    
    possible_values = list(set(X_test_array[:, sFeature.featureIndex])) # default = sFeature.unprivPop
    if len(possible_values) == 2:
        possible_values = possible_values[:-1]

    for value in possible_values: 

        results[str(value)] = []

        if sFeature.type == 0: # numeric sensitive feature
            sens_array = X_test_array[:, sFeature.featureIndex] 
            sens_array = binarize([sens_array.astype(float)],threshold=value)
        
        indexes0 = np.where(X_test_array[:, sFeature.featureIndex] == value)
        indexes1 = np.where(X_test_array[:, sFeature.featureIndex] != value) #sFeature.unprivPop
        
        y_test_0 = y_test_array[indexes0]
        y_test_1 = y_test_array[indexes1]
        y_pred_0 = y_pred[indexes0]
        y_pred_1 = y_pred[indexes1]

        # tn, fp, fn, tp = confusion_matrix(y_test, y_pred).ravel()
        cm0 = confusion_matrix(y_test_0, y_pred_0).ravel()
        cm1 = confusion_matrix(y_test_1, y_pred_1).ravel()

        if (len(cm0) == 4) and (len(cm1) == 4) :
            
            for i in metrics:
                if i == FairMetricEnum.DP:  
                    results[str(value)].append(demographic_parity(cm0,cm1))
                elif i == FairMetricEnum.EO:
                    results[str(value)].append(equal_opportunity(cm0,cm1))
                elif i == FairMetricEnum.PE:
                    results[str(value)].append(predictive_equality(cm0,cm1))
                elif i == FairMetricEnum.EOD:
                    results[str(value)].append(equalized_odds(cm0,cm1))
                elif i == FairMetricEnum.PP:
                    results[str(value)].append(predictive_parity(cm0,cm1))
                elif i == FairMetricEnum.CEA:
                    results[str(value)].append(conditional_accuracy_equality(cm0,cm1))
        
    return results



#
# Autogenerated by Thrift Compiler (0.18.1)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import logging
from .ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
all_structs = []


class Iface(object):
    def checkin(self, token, reportName):
        """
        Parameters:
         - token
         - reportName

        """
        pass

    def setSensitiveFeature(self, indexes):
        """
        Parameters:
         - indexes

        """
        pass

    def setModelDef(self, definitions):
        """
        Parameters:
         - definitions

        """
        pass

    def setTrainingData(self, list_data, y, list_features_name, nonnumeric_features):
        """
        Parameters:
         - list_data
         - y
         - list_features_name
         - nonnumeric_features

        """
        pass

    def setOriginalModel(self, model):
        """
        Parameters:
         - model

        """
        pass

    def setPredictions(self, labels):
        """
        Parameters:
         - labels

        """
        pass

    def setProbPredictions(self, labels):
        """
        Parameters:
         - labels

        """
        pass

    def checkout(self, sendMail):
        """
        Parameters:
         - sendMail

        """
        pass

    def setEvaluationMethod(self, evalMethod):
        """
        Parameters:
         - evalMethod

        """
        pass

    def getExplanations(self, isOriginalModel, dataInstance):
        """
        Parameters:
         - isOriginalModel
         - dataInstance

        """
        pass

    def getAExplanation(self, dataInstance):
        """
        Parameters:
         - dataInstance

        """
        pass

    def computeFairnessMetrics(self, isOriginalModel, metrics, sFeature, X_test, y_test):
        """
        Parameters:
         - isOriginalModel
         - metrics
         - sFeature
         - X_test
         - y_test

        """
        pass

    def debias(self):
        pass


class Client(Iface):
    def __init__(self, iprot, oprot=None):
        self._iprot = self._oprot = iprot
        if oprot is not None:
            self._oprot = oprot
        self._seqid = 0

    def checkin(self, token, reportName):
        """
        Parameters:
         - token
         - reportName

        """
        self.send_checkin(token, reportName)
        return self.recv_checkin()

    def send_checkin(self, token, reportName):
        self._oprot.writeMessageBegin('checkin', TMessageType.CALL, self._seqid)
        args = checkin_args()
        args.token = token
        args.reportName = reportName
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_checkin(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = checkin_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "checkin failed: unknown result")

    def setSensitiveFeature(self, indexes):
        """
        Parameters:
         - indexes

        """
        self.send_setSensitiveFeature(indexes)
        return self.recv_setSensitiveFeature()

    def send_setSensitiveFeature(self, indexes):
        self._oprot.writeMessageBegin('setSensitiveFeature', TMessageType.CALL, self._seqid)
        args = setSensitiveFeature_args()
        args.indexes = indexes
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_setSensitiveFeature(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = setSensitiveFeature_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "setSensitiveFeature failed: unknown result")

    def setModelDef(self, definitions):
        """
        Parameters:
         - definitions

        """
        self.send_setModelDef(definitions)
        return self.recv_setModelDef()

    def send_setModelDef(self, definitions):
        self._oprot.writeMessageBegin('setModelDef', TMessageType.CALL, self._seqid)
        args = setModelDef_args()
        args.definitions = definitions
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_setModelDef(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = setModelDef_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "setModelDef failed: unknown result")

    def setTrainingData(self, list_data, y, list_features_name, nonnumeric_features):
        """
        Parameters:
         - list_data
         - y
         - list_features_name
         - nonnumeric_features

        """
        self.send_setTrainingData(list_data, y, list_features_name, nonnumeric_features)
        return self.recv_setTrainingData()

    def send_setTrainingData(self, list_data, y, list_features_name, nonnumeric_features):
        self._oprot.writeMessageBegin('setTrainingData', TMessageType.CALL, self._seqid)
        args = setTrainingData_args()
        args.list_data = list_data
        args.y = y
        args.list_features_name = list_features_name
        args.nonnumeric_features = nonnumeric_features
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_setTrainingData(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = setTrainingData_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "setTrainingData failed: unknown result")

    def setOriginalModel(self, model):
        """
        Parameters:
         - model

        """
        self.send_setOriginalModel(model)
        return self.recv_setOriginalModel()

    def send_setOriginalModel(self, model):
        self._oprot.writeMessageBegin('setOriginalModel', TMessageType.CALL, self._seqid)
        args = setOriginalModel_args()
        args.model = model
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_setOriginalModel(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = setOriginalModel_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "setOriginalModel failed: unknown result")

    def setPredictions(self, labels):
        """
        Parameters:
         - labels

        """
        self.send_setPredictions(labels)
        return self.recv_setPredictions()

    def send_setPredictions(self, labels):
        self._oprot.writeMessageBegin('setPredictions', TMessageType.CALL, self._seqid)
        args = setPredictions_args()
        args.labels = labels
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_setPredictions(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = setPredictions_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "setPredictions failed: unknown result")

    def setProbPredictions(self, labels):
        """
        Parameters:
         - labels

        """
        self.send_setProbPredictions(labels)
        return self.recv_setProbPredictions()

    def send_setProbPredictions(self, labels):
        self._oprot.writeMessageBegin('setProbPredictions', TMessageType.CALL, self._seqid)
        args = setProbPredictions_args()
        args.labels = labels
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_setProbPredictions(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = setProbPredictions_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "setProbPredictions failed: unknown result")

    def checkout(self, sendMail):
        """
        Parameters:
         - sendMail

        """
        self.send_checkout(sendMail)
        return self.recv_checkout()

    def send_checkout(self, sendMail):
        self._oprot.writeMessageBegin('checkout', TMessageType.CALL, self._seqid)
        args = checkout_args()
        args.sendMail = sendMail
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_checkout(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = checkout_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "checkout failed: unknown result")

    def setEvaluationMethod(self, evalMethod):
        """
        Parameters:
         - evalMethod

        """
        self.send_setEvaluationMethod(evalMethod)
        return self.recv_setEvaluationMethod()

    def send_setEvaluationMethod(self, evalMethod):
        self._oprot.writeMessageBegin('setEvaluationMethod', TMessageType.CALL, self._seqid)
        args = setEvaluationMethod_args()
        args.evalMethod = evalMethod
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_setEvaluationMethod(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = setEvaluationMethod_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "setEvaluationMethod failed: unknown result")

    def getExplanations(self, isOriginalModel, dataInstance):
        """
        Parameters:
         - isOriginalModel
         - dataInstance

        """
        self.send_getExplanations(isOriginalModel, dataInstance)
        return self.recv_getExplanations()

    def send_getExplanations(self, isOriginalModel, dataInstance):
        self._oprot.writeMessageBegin('getExplanations', TMessageType.CALL, self._seqid)
        args = getExplanations_args()
        args.isOriginalModel = isOriginalModel
        args.dataInstance = dataInstance
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getExplanations(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getExplanations_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getExplanations failed: unknown result")

    def getAExplanation(self, dataInstance):
        """
        Parameters:
         - dataInstance

        """
        self.send_getAExplanation(dataInstance)
        return self.recv_getAExplanation()

    def send_getAExplanation(self, dataInstance):
        self._oprot.writeMessageBegin('getAExplanation', TMessageType.CALL, self._seqid)
        args = getAExplanation_args()
        args.dataInstance = dataInstance
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getAExplanation(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getAExplanation_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getAExplanation failed: unknown result")

    def computeFairnessMetrics(self, isOriginalModel, metrics, sFeature, X_test, y_test):
        """
        Parameters:
         - isOriginalModel
         - metrics
         - sFeature
         - X_test
         - y_test

        """
        self.send_computeFairnessMetrics(isOriginalModel, metrics, sFeature, X_test, y_test)
        return self.recv_computeFairnessMetrics()

    def send_computeFairnessMetrics(self, isOriginalModel, metrics, sFeature, X_test, y_test):
        self._oprot.writeMessageBegin('computeFairnessMetrics', TMessageType.CALL, self._seqid)
        args = computeFairnessMetrics_args()
        args.isOriginalModel = isOriginalModel
        args.metrics = metrics
        args.sFeature = sFeature
        args.X_test = X_test
        args.y_test = y_test
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_computeFairnessMetrics(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = computeFairnessMetrics_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "computeFairnessMetrics failed: unknown result")

    def debias(self):
        self.send_debias()
        return self.recv_debias()

    def send_debias(self):
        self._oprot.writeMessageBegin('debias', TMessageType.CALL, self._seqid)
        args = debias_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_debias(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = debias_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "debias failed: unknown result")


class Processor(Iface, TProcessor):
    def __init__(self, handler):
        self._handler = handler
        self._processMap = {}
        self._processMap["checkin"] = Processor.process_checkin
        self._processMap["setSensitiveFeature"] = Processor.process_setSensitiveFeature
        self._processMap["setModelDef"] = Processor.process_setModelDef
        self._processMap["setTrainingData"] = Processor.process_setTrainingData
        self._processMap["setOriginalModel"] = Processor.process_setOriginalModel
        self._processMap["setPredictions"] = Processor.process_setPredictions
        self._processMap["setProbPredictions"] = Processor.process_setProbPredictions
        self._processMap["checkout"] = Processor.process_checkout
        self._processMap["setEvaluationMethod"] = Processor.process_setEvaluationMethod
        self._processMap["getExplanations"] = Processor.process_getExplanations
        self._processMap["getAExplanation"] = Processor.process_getAExplanation
        self._processMap["computeFairnessMetrics"] = Processor.process_computeFairnessMetrics
        self._processMap["debias"] = Processor.process_debias
        self._on_message_begin = None

    def on_message_begin(self, func):
        self._on_message_begin = func

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if self._on_message_begin:
            self._on_message_begin(name, type, seqid)
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_checkin(self, seqid, iprot, oprot):
        args = checkin_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = checkin_result()
        try:
            result.success = self._handler.checkin(args.token, args.reportName)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("checkin", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_setSensitiveFeature(self, seqid, iprot, oprot):
        args = setSensitiveFeature_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = setSensitiveFeature_result()
        try:
            result.success = self._handler.setSensitiveFeature(args.indexes)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("setSensitiveFeature", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_setModelDef(self, seqid, iprot, oprot):
        args = setModelDef_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = setModelDef_result()
        try:
            result.success = self._handler.setModelDef(args.definitions)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("setModelDef", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_setTrainingData(self, seqid, iprot, oprot):
        args = setTrainingData_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = setTrainingData_result()
        try:
            result.success = self._handler.setTrainingData(args.list_data, args.y, args.list_features_name, args.nonnumeric_features)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("setTrainingData", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_setOriginalModel(self, seqid, iprot, oprot):
        args = setOriginalModel_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = setOriginalModel_result()
        try:
            result.success = self._handler.setOriginalModel(args.model)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("setOriginalModel", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_setPredictions(self, seqid, iprot, oprot):
        args = setPredictions_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = setPredictions_result()
        try:
            result.success = self._handler.setPredictions(args.labels)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("setPredictions", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_setProbPredictions(self, seqid, iprot, oprot):
        args = setProbPredictions_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = setProbPredictions_result()
        try:
            result.success = self._handler.setProbPredictions(args.labels)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("setProbPredictions", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_checkout(self, seqid, iprot, oprot):
        args = checkout_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = checkout_result()
        try:
            result.success = self._handler.checkout(args.sendMail)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("checkout", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_setEvaluationMethod(self, seqid, iprot, oprot):
        args = setEvaluationMethod_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = setEvaluationMethod_result()
        try:
            result.success = self._handler.setEvaluationMethod(args.evalMethod)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("setEvaluationMethod", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getExplanations(self, seqid, iprot, oprot):
        args = getExplanations_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getExplanations_result()
        try:
            result.success = self._handler.getExplanations(args.isOriginalModel, args.dataInstance)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getExplanations", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getAExplanation(self, seqid, iprot, oprot):
        args = getAExplanation_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getAExplanation_result()
        try:
            result.success = self._handler.getAExplanation(args.dataInstance)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getAExplanation", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_computeFairnessMetrics(self, seqid, iprot, oprot):
        args = computeFairnessMetrics_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = computeFairnessMetrics_result()
        try:
            result.success = self._handler.computeFairnessMetrics(args.isOriginalModel, args.metrics, args.sFeature, args.X_test, args.y_test)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("computeFairnessMetrics", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_debias(self, seqid, iprot, oprot):
        args = debias_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = debias_result()
        try:
            result.success = self._handler.debias()
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("debias", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

# HELPER FUNCTIONS AND STRUCTURES


class checkin_args(object):
    """
    Attributes:
     - token
     - reportName

    """


    def __init__(self, token=None, reportName=None,):
        self.token = token
        self.reportName = reportName

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.token = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.reportName = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('checkin_args')
        if self.token is not None:
            oprot.writeFieldBegin('token', TType.STRING, 1)
            oprot.writeString(self.token.encode('utf-8') if sys.version_info[0] == 2 else self.token)
            oprot.writeFieldEnd()
        if self.reportName is not None:
            oprot.writeFieldBegin('reportName', TType.STRING, 2)
            oprot.writeString(self.reportName.encode('utf-8') if sys.version_info[0] == 2 else self.reportName)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(checkin_args)
checkin_args.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'token', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'reportName', 'UTF8', None, ),  # 2
)


class checkin_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.BOOL:
                    self.success = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('checkin_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.BOOL, 0)
            oprot.writeBool(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(checkin_result)
checkin_result.thrift_spec = (
    (0, TType.BOOL, 'success', None, None, ),  # 0
)


class setSensitiveFeature_args(object):
    """
    Attributes:
     - indexes

    """


    def __init__(self, indexes=None,):
        self.indexes = indexes

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.indexes = []
                    (_etype17, _size14) = iprot.readListBegin()
                    for _i18 in range(_size14):
                        _elem19 = SensitiveFeature()
                        _elem19.read(iprot)
                        self.indexes.append(_elem19)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setSensitiveFeature_args')
        if self.indexes is not None:
            oprot.writeFieldBegin('indexes', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.indexes))
            for iter20 in self.indexes:
                iter20.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setSensitiveFeature_args)
setSensitiveFeature_args.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'indexes', (TType.STRUCT, [SensitiveFeature, None], False), None, ),  # 1
)


class setSensitiveFeature_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.LIST:
                    self.success = []
                    (_etype24, _size21) = iprot.readListBegin()
                    for _i25 in range(_size21):
                        _elem26 = SensitiveFeature()
                        _elem26.read(iprot)
                        self.success.append(_elem26)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setSensitiveFeature_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.LIST, 0)
            oprot.writeListBegin(TType.STRUCT, len(self.success))
            for iter27 in self.success:
                iter27.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setSensitiveFeature_result)
setSensitiveFeature_result.thrift_spec = (
    (0, TType.LIST, 'success', (TType.STRUCT, [SensitiveFeature, None], False), None, ),  # 0
)


class setModelDef_args(object):
    """
    Attributes:
     - definitions

    """


    def __init__(self, definitions=None,):
        self.definitions = definitions

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.definitions = ModelDefinitions()
                    self.definitions.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setModelDef_args')
        if self.definitions is not None:
            oprot.writeFieldBegin('definitions', TType.STRUCT, 1)
            self.definitions.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setModelDef_args)
setModelDef_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'definitions', [ModelDefinitions, None], None, ),  # 1
)


class setModelDef_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.I32:
                    self.success = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setModelDef_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.I32, 0)
            oprot.writeI32(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setModelDef_result)
setModelDef_result.thrift_spec = (
    (0, TType.I32, 'success', None, None, ),  # 0
)


class setTrainingData_args(object):
    """
    Attributes:
     - list_data
     - y
     - list_features_name
     - nonnumeric_features

    """


    def __init__(self, list_data=None, y=None, list_features_name=None, nonnumeric_features=None,):
        self.list_data = list_data
        self.y = y
        self.list_features_name = list_features_name
        self.nonnumeric_features = nonnumeric_features

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.list_data = []
                    (_etype31, _size28) = iprot.readListBegin()
                    for _i32 in range(_size28):
                        _elem33 = []
                        (_etype37, _size34) = iprot.readListBegin()
                        for _i38 in range(_size34):
                            _elem39 = iprot.readDouble()
                            _elem33.append(_elem39)
                        iprot.readListEnd()
                        self.list_data.append(_elem33)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.y = []
                    (_etype43, _size40) = iprot.readListBegin()
                    for _i44 in range(_size40):
                        _elem45 = iprot.readDouble()
                        self.y.append(_elem45)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.list_features_name = []
                    (_etype49, _size46) = iprot.readListBegin()
                    for _i50 in range(_size46):
                        _elem51 = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                        self.list_features_name.append(_elem51)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.nonnumeric_features = []
                    (_etype55, _size52) = iprot.readListBegin()
                    for _i56 in range(_size52):
                        _elem57 = iprot.readI32()
                        self.nonnumeric_features.append(_elem57)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setTrainingData_args')
        if self.list_data is not None:
            oprot.writeFieldBegin('list_data', TType.LIST, 1)
            oprot.writeListBegin(TType.LIST, len(self.list_data))
            for iter58 in self.list_data:
                oprot.writeListBegin(TType.DOUBLE, len(iter58))
                for iter59 in iter58:
                    oprot.writeDouble(iter59)
                oprot.writeListEnd()
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.y is not None:
            oprot.writeFieldBegin('y', TType.LIST, 2)
            oprot.writeListBegin(TType.DOUBLE, len(self.y))
            for iter60 in self.y:
                oprot.writeDouble(iter60)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.list_features_name is not None:
            oprot.writeFieldBegin('list_features_name', TType.LIST, 3)
            oprot.writeListBegin(TType.STRING, len(self.list_features_name))
            for iter61 in self.list_features_name:
                oprot.writeString(iter61.encode('utf-8') if sys.version_info[0] == 2 else iter61)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.nonnumeric_features is not None:
            oprot.writeFieldBegin('nonnumeric_features', TType.LIST, 4)
            oprot.writeListBegin(TType.I32, len(self.nonnumeric_features))
            for iter62 in self.nonnumeric_features:
                oprot.writeI32(iter62)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setTrainingData_args)
setTrainingData_args.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'list_data', (TType.LIST, (TType.DOUBLE, None, False), False), None, ),  # 1
    (2, TType.LIST, 'y', (TType.DOUBLE, None, False), None, ),  # 2
    (3, TType.LIST, 'list_features_name', (TType.STRING, 'UTF8', False), None, ),  # 3
    (4, TType.LIST, 'nonnumeric_features', (TType.I32, None, False), None, ),  # 4
)


class setTrainingData_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.I32:
                    self.success = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setTrainingData_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.I32, 0)
            oprot.writeI32(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setTrainingData_result)
setTrainingData_result.thrift_spec = (
    (0, TType.I32, 'success', None, None, ),  # 0
)


class setOriginalModel_args(object):
    """
    Attributes:
     - model

    """


    def __init__(self, model=None,):
        self.model = model

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.model = iprot.readBinary()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setOriginalModel_args')
        if self.model is not None:
            oprot.writeFieldBegin('model', TType.STRING, 1)
            oprot.writeBinary(self.model)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setOriginalModel_args)
setOriginalModel_args.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'model', 'BINARY', None, ),  # 1
)


class setOriginalModel_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.I32:
                    self.success = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setOriginalModel_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.I32, 0)
            oprot.writeI32(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setOriginalModel_result)
setOriginalModel_result.thrift_spec = (
    (0, TType.I32, 'success', None, None, ),  # 0
)


class setPredictions_args(object):
    """
    Attributes:
     - labels

    """


    def __init__(self, labels=None,):
        self.labels = labels

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.labels = []
                    (_etype66, _size63) = iprot.readListBegin()
                    for _i67 in range(_size63):
                        _elem68 = iprot.readDouble()
                        self.labels.append(_elem68)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setPredictions_args')
        if self.labels is not None:
            oprot.writeFieldBegin('labels', TType.LIST, 1)
            oprot.writeListBegin(TType.DOUBLE, len(self.labels))
            for iter69 in self.labels:
                oprot.writeDouble(iter69)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setPredictions_args)
setPredictions_args.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'labels', (TType.DOUBLE, None, False), None, ),  # 1
)


class setPredictions_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.I32:
                    self.success = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setPredictions_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.I32, 0)
            oprot.writeI32(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setPredictions_result)
setPredictions_result.thrift_spec = (
    (0, TType.I32, 'success', None, None, ),  # 0
)


class setProbPredictions_args(object):
    """
    Attributes:
     - labels

    """


    def __init__(self, labels=None,):
        self.labels = labels

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.labels = []
                    (_etype73, _size70) = iprot.readListBegin()
                    for _i74 in range(_size70):
                        _elem75 = iprot.readDouble()
                        self.labels.append(_elem75)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setProbPredictions_args')
        if self.labels is not None:
            oprot.writeFieldBegin('labels', TType.LIST, 1)
            oprot.writeListBegin(TType.DOUBLE, len(self.labels))
            for iter76 in self.labels:
                oprot.writeDouble(iter76)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setProbPredictions_args)
setProbPredictions_args.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'labels', (TType.DOUBLE, None, False), None, ),  # 1
)


class setProbPredictions_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.I32:
                    self.success = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setProbPredictions_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.I32, 0)
            oprot.writeI32(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setProbPredictions_result)
setProbPredictions_result.thrift_spec = (
    (0, TType.I32, 'success', None, None, ),  # 0
)


class checkout_args(object):
    """
    Attributes:
     - sendMail

    """


    def __init__(self, sendMail=None,):
        self.sendMail = sendMail

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.BOOL:
                    self.sendMail = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('checkout_args')
        if self.sendMail is not None:
            oprot.writeFieldBegin('sendMail', TType.BOOL, 1)
            oprot.writeBool(self.sendMail)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(checkout_args)
checkout_args.thrift_spec = (
    None,  # 0
    (1, TType.BOOL, 'sendMail', None, None, ),  # 1
)


class checkout_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRING:
                    self.success = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('checkout_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRING, 0)
            oprot.writeString(self.success.encode('utf-8') if sys.version_info[0] == 2 else self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(checkout_result)
checkout_result.thrift_spec = (
    (0, TType.STRING, 'success', 'UTF8', None, ),  # 0
)


class setEvaluationMethod_args(object):
    """
    Attributes:
     - evalMethod

    """


    def __init__(self, evalMethod=None,):
        self.evalMethod = evalMethod

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.evalMethod = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setEvaluationMethod_args')
        if self.evalMethod is not None:
            oprot.writeFieldBegin('evalMethod', TType.I32, 1)
            oprot.writeI32(self.evalMethod)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setEvaluationMethod_args)
setEvaluationMethod_args.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'evalMethod', None, None, ),  # 1
)


class setEvaluationMethod_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.I32:
                    self.success = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('setEvaluationMethod_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.I32, 0)
            oprot.writeI32(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(setEvaluationMethod_result)
setEvaluationMethod_result.thrift_spec = (
    (0, TType.I32, 'success', None, None, ),  # 0
)


class getExplanations_args(object):
    """
    Attributes:
     - isOriginalModel
     - dataInstance

    """


    def __init__(self, isOriginalModel=None, dataInstance=None,):
        self.isOriginalModel = isOriginalModel
        self.dataInstance = dataInstance

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.BOOL:
                    self.isOriginalModel = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.dataInstance = []
                    (_etype80, _size77) = iprot.readListBegin()
                    for _i81 in range(_size77):
                        _elem82 = []
                        (_etype86, _size83) = iprot.readListBegin()
                        for _i87 in range(_size83):
                            _elem88 = iprot.readDouble()
                            _elem82.append(_elem88)
                        iprot.readListEnd()
                        self.dataInstance.append(_elem82)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getExplanations_args')
        if self.isOriginalModel is not None:
            oprot.writeFieldBegin('isOriginalModel', TType.BOOL, 1)
            oprot.writeBool(self.isOriginalModel)
            oprot.writeFieldEnd()
        if self.dataInstance is not None:
            oprot.writeFieldBegin('dataInstance', TType.LIST, 2)
            oprot.writeListBegin(TType.LIST, len(self.dataInstance))
            for iter89 in self.dataInstance:
                oprot.writeListBegin(TType.DOUBLE, len(iter89))
                for iter90 in iter89:
                    oprot.writeDouble(iter90)
                oprot.writeListEnd()
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getExplanations_args)
getExplanations_args.thrift_spec = (
    None,  # 0
    (1, TType.BOOL, 'isOriginalModel', None, None, ),  # 1
    (2, TType.LIST, 'dataInstance', (TType.LIST, (TType.DOUBLE, None, False), False), None, ),  # 2
)


class getExplanations_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = FIExplanation()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getExplanations_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getExplanations_result)
getExplanations_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [FIExplanation, None], None, ),  # 0
)


class getAExplanation_args(object):
    """
    Attributes:
     - dataInstance

    """


    def __init__(self, dataInstance=None,):
        self.dataInstance = dataInstance

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.dataInstance = []
                    (_etype94, _size91) = iprot.readListBegin()
                    for _i95 in range(_size91):
                        _elem96 = iprot.readDouble()
                        self.dataInstance.append(_elem96)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getAExplanation_args')
        if self.dataInstance is not None:
            oprot.writeFieldBegin('dataInstance', TType.LIST, 1)
            oprot.writeListBegin(TType.DOUBLE, len(self.dataInstance))
            for iter97 in self.dataInstance:
                oprot.writeDouble(iter97)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getAExplanation_args)
getAExplanation_args.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'dataInstance', (TType.DOUBLE, None, False), None, ),  # 1
)


class getAExplanation_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = FIExplanation()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getAExplanation_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getAExplanation_result)
getAExplanation_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [FIExplanation, None], None, ),  # 0
)


class computeFairnessMetrics_args(object):
    """
    Attributes:
     - isOriginalModel
     - metrics
     - sFeature
     - X_test
     - y_test

    """


    def __init__(self, isOriginalModel=None, metrics=None, sFeature=None, X_test=None, y_test=None,):
        self.isOriginalModel = isOriginalModel
        self.metrics = metrics
        self.sFeature = sFeature
        self.X_test = X_test
        self.y_test = y_test

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.BOOL:
                    self.isOriginalModel = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.metrics = []
                    (_etype101, _size98) = iprot.readListBegin()
                    for _i102 in range(_size98):
                        _elem103 = iprot.readI32()
                        self.metrics.append(_elem103)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRUCT:
                    self.sFeature = SensitiveFeature()
                    self.sFeature.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.X_test = []
                    (_etype107, _size104) = iprot.readListBegin()
                    for _i108 in range(_size104):
                        _elem109 = []
                        (_etype113, _size110) = iprot.readListBegin()
                        for _i114 in range(_size110):
                            _elem115 = iprot.readDouble()
                            _elem109.append(_elem115)
                        iprot.readListEnd()
                        self.X_test.append(_elem109)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.LIST:
                    self.y_test = []
                    (_etype119, _size116) = iprot.readListBegin()
                    for _i120 in range(_size116):
                        _elem121 = iprot.readI32()
                        self.y_test.append(_elem121)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('computeFairnessMetrics_args')
        if self.isOriginalModel is not None:
            oprot.writeFieldBegin('isOriginalModel', TType.BOOL, 1)
            oprot.writeBool(self.isOriginalModel)
            oprot.writeFieldEnd()
        if self.metrics is not None:
            oprot.writeFieldBegin('metrics', TType.LIST, 2)
            oprot.writeListBegin(TType.I32, len(self.metrics))
            for iter122 in self.metrics:
                oprot.writeI32(iter122)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.sFeature is not None:
            oprot.writeFieldBegin('sFeature', TType.STRUCT, 3)
            self.sFeature.write(oprot)
            oprot.writeFieldEnd()
        if self.X_test is not None:
            oprot.writeFieldBegin('X_test', TType.LIST, 4)
            oprot.writeListBegin(TType.LIST, len(self.X_test))
            for iter123 in self.X_test:
                oprot.writeListBegin(TType.DOUBLE, len(iter123))
                for iter124 in iter123:
                    oprot.writeDouble(iter124)
                oprot.writeListEnd()
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.y_test is not None:
            oprot.writeFieldBegin('y_test', TType.LIST, 5)
            oprot.writeListBegin(TType.I32, len(self.y_test))
            for iter125 in self.y_test:
                oprot.writeI32(iter125)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(computeFairnessMetrics_args)
computeFairnessMetrics_args.thrift_spec = (
    None,  # 0
    (1, TType.BOOL, 'isOriginalModel', None, None, ),  # 1
    (2, TType.LIST, 'metrics', (TType.I32, None, False), None, ),  # 2
    (3, TType.STRUCT, 'sFeature', [SensitiveFeature, None], None, ),  # 3
    (4, TType.LIST, 'X_test', (TType.LIST, (TType.DOUBLE, None, False), False), None, ),  # 4
    (5, TType.LIST, 'y_test', (TType.I32, None, False), None, ),  # 5
)


class computeFairnessMetrics_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.LIST:
                    self.success = []
                    (_etype129, _size126) = iprot.readListBegin()
                    for _i130 in range(_size126):
                        _elem131 = FairMetric()
                        _elem131.read(iprot)
                        self.success.append(_elem131)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('computeFairnessMetrics_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.LIST, 0)
            oprot.writeListBegin(TType.STRUCT, len(self.success))
            for iter132 in self.success:
                iter132.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(computeFairnessMetrics_result)
computeFairnessMetrics_result.thrift_spec = (
    (0, TType.LIST, 'success', (TType.STRUCT, [FairMetric, None], False), None, ),  # 0
)


class debias_args(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('debias_args')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(debias_args)
debias_args.thrift_spec = (
)


class debias_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRING:
                    self.success = iprot.readBinary()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('debias_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRING, 0)
            oprot.writeBinary(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(debias_result)
debias_result.thrift_spec = (
    (0, TType.STRING, 'success', 'BINARY', None, ),  # 0
)
fix_spec(all_structs)
del all_structs



from setuptools import setup, find_packages

setup(

    name='fixout',
    version='0.1.29',
    description='Algorithmic inspection for trustworthy ML models',
    packages=find_packages(),
    package_data={'': ['web/templates/*.html',
                       'web/templates/utils-web/*.html',
                       'web/static/*.css',
                       'web/static/*.png',
                       'web/static/*.svg',
                       'demos/*.py',
                       'demos/data/*.data']},
    include_package_data=True,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/fixouttech/fixout",
    author="FixOut",
    license="BSD",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.2",
        "Operating System :: OS Independent"
    ],
    python_requires=">=3.8",
    install_requires=[
        'Flask>=3.0.3',
        'Jinja2>=3.1.4',
        'numpy>=1.26.4',
        'pandas>=2.2.2',
        'plotly>=5.24.0',
        'scikit-image>=0.24.0',
        'scikit-learn>=1.5.2',
        'scipy>=1.11.4',
        'thrift>=0.20.0',
        'tinycss2>=1.4.0',
        'webcolors>=24.11.1'
    ],
)